/*
 * Program to test how fast and reliably bus read/writes can be done using a 
 * simple synchronous parallel bus connecting two or more Propellers. 
 *
 * There can be only one sender cog at a time, but can be multiple receiver 
 * cogs. The intention is to start one sender cog on one Propeller, and 
 * one or more receiver cogs on other Propellers. This program supports
 * up to four receivers on a single Propeller. In sender mode the program
 * just sends forever, and in receive mode it just receives forever, but
 * in transceiver mode the program swaps between sending and receiving, 
 * which is a more realistic scenario.
 *
 * The Propellers must be connected pin to pin on pins 00 .. 31 (i.e. port A). 
 * If you have more than two Propellers connected, you could start additional 
 * receiver cogs on the other Propellers.
 *
 * This is a synchronous transfer, so how much data can be transferred depends
 * on how closely the respective Propeller clocks are synchronized. The crystal 
 * accuracy is typically +/-0.5 PPM. At 25 clocks per long, this means the 
 * clocks may be out of sync by up one clock after 40,000 longs. A suitable 
 * maximum size for a single synchronous transfer might therefore be 10,000,
 * but may be increased depending on the accuracy of the Propeller clocks. 
 * Larger transfers can be performed by doing multiple smaller transfers 
 * (which is the point of this test program!).
 *
 * The program has no synchronization mechanism, except that the receiver
 * looks for the value 0x00000001 to indcate the start of a transfer. This
 * is useful when the program is compiled as both sender and receiver, to
 * prevent the programs getting locked in a bus contention situation where 
 * both ends keep transmitting at the same time. Eventually, a higher level 
 * protocol and/or a 'bus master' will be used to resolve bus contention.
 *
 * The program uses P2 NATIVE PASM, so it must be compiled in P2 NATIVE mode
 * (which is the default mode for the Propeller 2). 
 *
 * To maximize available cogs, add -C NO_MOUSE and -C NO_FLOAT (and -C SIMPLE
 * if using a serial HMI). However in this simple-minded program, buffer space
 * is likely to be the limiting factor, not cogs.
 *
 *  To build as a sender, add -C SENDER
 *
 *  To build as a receiver, add -C RECEIVER
 *
 *  To build as a transceiver (i.e. both a sender and a receiver) 
 *  add -C RECEIVER -C SENDER
 *
 * For example, compile with a command like:
 *
 *    catalina -p2 -lci p2p.c -o sender -C NO_MOUSE -C NO_FLOAT -C SENDER
 * or
 *    catalina -p2 -lci p2p.c -o receiver -C NO_MOUSE -C NO_FLOAT -C RECEIVER
 *
 *    catalina -p2 -lci p2p.c -o transceiver -C NO_MOUSE -C NO_FLOAT -C RECEIVER -C SENDER
 *
 * Then load and execute the sender and receiver with commands like:
 *
 *    payload sender -PN -i
 *    payload receiver -PM -i
 *
 * where N & M are the Propeller ports to use for the sender and receiver,
 * respectively (add more 'receiver' commands if there are more than two
 * Propellers on the bus).
 *
 * Or load and execute the transceiver with commands like:
 *
 *    payload transceiver -PN -i
 *    payload transceiver -PM -i
 *
 * Note that in transceiver mode, only two instances of the program can
 * be executed at once.
 *
 */

#if !defined (__CATALINA_SENDER) && !defined(__CATALINA_RECEIVER)
#error AT LEAST ONE OF SENDER OR RECEIVER MUST BE DEFINED!
#endif

#include <stdio.h>
#include <stdlib.h>
#include <catalina_cog.h>
#include <catalina_plugin.h>

#define XFER_DELAY 7             // clocks per long is 18+this (7 works!)

#define NUM_LONGS  10000         // number of longs to transfer (10000 work!)

#define STACK_SIZE 500           // size of cog stack (stdio requires 500)

#define TOTAL_ERRORS_ONLY        // print only total errors, not details

static unsigned long start = 0;  // clock count used to synchronize cogs
static int lock = 0;             // lock to protect I/O

static unsigned long *send_buff; // data to be sent
static unsigned long *rcv1_buff; // data received (1)
static unsigned long *rcv2_buff; // data received (2)
static unsigned long *rcv3_buff; // data received (3)
static unsigned long *rcv4_buff; // data received (4)

/*
 * sync - synchronize multiple cogs to start on a specific clock count
 *
 *    'start' should be set to a clock count some time in the 
 *            future - e.g. _cnt() + _clockfreq() for one second
 */
int sync(unsigned long start) {
   return PASM (
      " getct  r0\n"
      " sub    r2, r0\n"
      " waitx  r2\n"
      " getct  r0\n"
   );
}

/*
 * send - pasm code to write a number of longs to the 32 bit bus
 *        Note: sending starts immediately, and then sends a new long
 *              every 'time'+18 ticks.
 *        Note: loads the code into LUT RAM and executes it there.
 *
 *    'time' (passed in r4) is the time between writes in clocks (+18)
 *    'buff' (passed in r3) is an array holding longs to send
 *    'size' (passed in r2) is the size of the array
 */
int send(int time, void *buff, int size) {
   return PASM (
      // load LUT RAM:
      "        setq2   #(send_end - send_start - 1)\n"
      "        rdlong  0, ##@send_start\n"

      // jump to code in LUT RAM:
      "        jmp     #send_start\n" 

      // code to be executed in LUT RAM:
      "        org     $200\n"
      "send_start\n"
      "        getct   r0\n"             // LUT: 2 (clocks)
      "\n"
      "        rep     #7, r2\n"         // LUT: 2
      "        addct1  r0, #18\n"        // LUT: 2
      "        rdlong  r1, r3\n"         // LUT: 9 .. 16
      "        waitct1\n"                // LUT: 2
      "        add     r3, #4\n"         // LUT: 2
      "        mov     outa, r1\n"       // LUT: 2
      "        addct1  r0, r4\n"         // LUT: 2
      "        waitct1\n"                // LUT: 2
      "\n"
      "        addct1  r0, r4\n"
      "        addct1  r0, #18\n"
      "        waitct1\n"
      "        mov     outa, #0\n"
      "        getct   r0\n"
      "        jmp     #send_cont\n"
      "send_end\n" 

      // resume Hub Execution:
      "       orgh\n"
      "send_cont\n" 
   );
}

/*
 * sender - send an array of longs to one or more receivers
 *
 *    'buff' is an array of NUM_LONGS longs.
 */
void sender(void *buff) {
   unsigned int started, stopped, total;
   int me = _cogid();

   // set pins 00 .. 31 as outputs
   PASM(" mov outa, #0\n"
        " or  dira, ##$FFFFFFFF\n");

   started = _cnt();
   stopped = send(XFER_DELAY, buff, NUM_LONGS);
   total   = stopped - started;
   ACQUIRE(lock);
   printf("send (cog %d) took %d clocks (%d per long)\n", 
           me, total, total/NUM_LONGS);
   RELEASE(lock);

   // set pins 00 .. 31 as inputs
   PASM(" andn dira, ##$FFFFFFFF\n");

   while(1); // don't exit
}

/*
 * recv - pasm code to read a number of longs from the 32 bit bus. 
 *        Note: receiving starts 'time' clock ticks after any non-zero
 *              value is detected on the bus, and then reads a new long 
 *              every 'time'+18 ticks.
 *        Note: loads the code into LUT RAM and executes it there.
 *
 *    'time' (passed in r4) is the time between reads in clocks (+18)
 *    'buff' (passed in r3) is an array to hold longs received
 *    'size' (passed in r2) is the size of the array
 */
int recv(int time, void *buff, int size) {
   return PASM (
      // load LUT RAM:
      "        setq2   #(recv_end - recv_start - 1)\n"
      "        rdlong  0, ##@recv_start\n"

      // jump to code in LUT RAM:
      "        jmp     #recv_start\n"

      // code to be executed in LUT RAM:
      "        org     $200\n"
      "recv_start\n"
      "        mov     r1, ina\n"        // LUT: 2 (clocks)
      "        cmp     r1, #1 wz\n"      // LUT: 2
      " if_nz   jmp     #recv_start\n"    // LUT: 4
      "        getct   r0\n"             // LUT: 2 
      "\n"
      "        rep     #7, r2\n"         // LUT: 2
      "        addct1  r0, #18\n"        // LUT: 2
      "        wrlong  r1, r3\n"         // LUT: 3 .. 10
      "        waitct1\n"                // LUT: 2
      "        add     r3, #4\n"         // LUT: 2
      "        addct1  r0, r4\n"         // LUT: 2
      "        waitct1\n"                // LUT: 2
      "        mov     r1, ina\n"        // LUT: 2 
      "\n"
      "        getct   r0\n"
      "        jmp     #recv_cont\n"
      "recv_end\n"

      // resume Hub Execution:
      "        orgh\n"
      "recv_cont\n"
   );
}

/*
 * receiver- receive an array of longs from the 32 bit bus
 *
 *    'buff' is an array of NUM_LONGS longs.
 */
void receiver(void *buff) {
   unsigned int started, stopped;
   int me = _cogid();

   // set pins 00 .. 31 as inputs
   PASM(" andn dira, ##$FFFFFFFF\n");

#if !(defined(__CATALINA_RECEIVER) && defined(__CATALINA_SENDER))
   // not transceiver mode, so syncing multiple receivers is desirable
   started = sync(start);
   stopped = recv(XFER_DELAY, buff, NUM_LONGS);
   ACQUIRE(lock);
   printf("recv (cog %d) started at clock 0x%08x\n", me, started);
   RELEASE(lock);
#else
   // in transceiver mode, syncing is not necessary
   recv(XFER_DELAY, buff, NUM_LONGS);
#endif

   while(1); // don't exit
}


void main(void) {
   unsigned long i;
   long send_stack[STACK_SIZE];
   long rcv1_stack[STACK_SIZE];
   long rcv2_stack[STACK_SIZE];
   long rcv3_stack[STACK_SIZE];
   long rcv4_stack[STACK_SIZE];
   int send_cog;
   int rcv1_cog;
   int rcv2_cog;
   int rcv3_cog;
   int rcv4_cog;
   int errors = 0;
   int transfers = 0;
   char ch;

   // assign a lock to be used to avoid plugin contention
   lock = _locknew();

   // give the vt100 emulator (if used) a chance to start
#ifdef __CATALINA_VT100
   _waitms(500);
#endif

   // allocate the arrays (we allocate them all whether we are a sender
   // or a receiver)
   send_buff = malloc(NUM_LONGS*4);
   rcv1_buff = malloc(NUM_LONGS*4);
   rcv2_buff = malloc(NUM_LONGS*4);
   rcv3_buff = malloc(NUM_LONGS*4);
   rcv4_buff = malloc(NUM_LONGS*4);

   // initialize the arrays
   for (i = 0; i < NUM_LONGS; i++) {
      send_buff[i] = i+1; // can be anything, but must be non-zero
      rcv1_buff[i] = 0;
      rcv2_buff[i] = 0;
      rcv3_buff[i] = 0;
      rcv4_buff[i] = 0;
   }

#if defined(__CATALINA_SENDER) && !defined(__CATALINA_RECEIVER)

   k_clear();
   ACQUIRE(lock);
   printf("SENDER (Clock = %lu Hz)\n\n", _clockfreq());
   printf("Start the receiver cogs in the Receiver program,\n");
   printf("then press any key to start sender cog\n");
   RELEASE(lock);
   k_wait();

   // keep transferring forever
   while (1) {
      ACQUIRE(lock);
      printf("Starting sender cog ...\n");
      RELEASE(lock);

      // start ONE sender
      send_cog = _cogstart_C(&sender, send_buff, send_stack, STACK_SIZE);

      // give the sender a chance to send (3 seconds is generous!)
      _waitms(3000);
      ACQUIRE(lock);
      printf("... done\n");
      RELEASE(lock);

      // cancel the sender (we restart it again for each transfer)
      _cogstop(send_cog);

      // update and print statistics
      transfers++;
      ACQUIRE(lock);
      printf("\nTotal %d transfers\n",transfers); 
      RELEASE(lock);
   }

#endif

#if defined(__CATALINA_RECEIVER) && !defined(__CATALINA_SENDER)

   ACQUIRE(lock);
   printf("RECEIVER (Clock = %lu Hz)\n\n", _clockfreq());
   printf("Press any key to start receiver cogs, then start the sender\n");
   printf("cog in the sender program\n");
   RELEASE(lock);
   k_wait();

   // keep transferring forever
   while (1) {
      ACQUIRE(lock);
      printf("Starting receiver cogs ...\n");
      RELEASE(lock);
      // set a start time for the receiver cogs to use in the sync function
      start = _cnt() + _clockfreq(); // set start time for +1 seconds

      // start FOUR receiver cogs
      rcv1_cog = _cogstart_C(&receiver, rcv1_buff, rcv1_stack, STACK_SIZE);
      rcv2_cog = _cogstart_C(&receiver, rcv2_buff, rcv2_stack, STACK_SIZE);
      rcv3_cog = _cogstart_C(&receiver, rcv3_buff, rcv3_stack, STACK_SIZE);
      rcv4_cog = _cogstart_C(&receiver, rcv4_buff, rcv4_stack, STACK_SIZE);
      ACQUIRE(lock);
      printf("... done\n");
      RELEASE(lock);

      // give receiver a chance to receive
      _waitms(1000);

      // wait till receivers complete
      while (
          (rcv1_buff[NUM_LONGS - 1] == 0)
       && (rcv2_buff[NUM_LONGS - 1] == 0)
       && (rcv3_buff[NUM_LONGS - 1] == 0)
       && (rcv4_buff[NUM_LONGS - 1] == 0)
      ) {
          _waitms(1000);
      }

      // terminate the receiver cogs (we restart them again for each transfer)
      _cogstop(rcv1_cog);
      _cogstop(rcv2_cog);
      _cogstop(rcv3_cog);
      _cogstop(rcv4_cog);

      // check the results
      ACQUIRE(lock);
      printf("\nChecking data ...\n");
      for (i = 0; i < NUM_LONGS; i++) {
         // check receiver 1 got the correct data
         if (send_buff[i] != rcv1_buff[i]) {
            printf("send[%3d]=0x%08X != rcv1[%3d]=0x%08X\n",  
                   i, send_buff[i], i, rcv1_buff[i]);
            _waitms(5);
#ifdef TOTAL_ERRORS_ONLY
            // if only counting total errors, we are done - we don't 
            // report each mismatch
            errors++;
            break;
#endif
         }
         // check receiver 2 got the correct data
         if (send_buff[i] != rcv2_buff[i]) {
            printf("send[%3d]=0x%08X != rcv2[%3d]=0x%08X\n",  
                   i, send_buff[i], i, rcv2_buff[i]);
            _waitms(5);
#ifdef TOTAL_ERRORS_ONLY
            // if only counting total errors, we are done - we don't 
            // report each mismatch
            errors++;
            break;
#endif
         }
         // check receiver 3 got the correct data
         if (send_buff[i] != rcv3_buff[i]) {
            printf("send[%3d]=0x%08X != rcv3[%3d]=0x%08X\n",  
                   i, send_buff[i], i, rcv3_buff[i]);
            _waitms(5);
#ifdef TOTAL_ERRORS_ONLY
            // if only counting total errors, we are done - we don't 
            // report each mismatch
            errors++;
            break;
#endif
         }
         // check receiver 4 got the correct data
         if (send_buff[i] != rcv4_buff[i]) {
            printf("send[%3d]=0x%08X != rcv4[%3d]=0x%08X\n",  
                   i, send_buff[i], i, rcv4_buff[i]);
            _waitms(5);
#ifdef TOTAL_ERRORS_ONLY
            // if only counting total errors, we are done - we don't 
            // report each mismatch
            errors++;
            break;
#endif
         }
      }
      printf("... done\n");

      // update and print statistics
      transfers++;
      printf("Total errors = %d (from %d transfers)\n\n", errors, transfers); 
      RELEASE(lock);

      // re-initialize the arrays for the next transfer
      for (i = 0; i < NUM_LONGS; i++) {
         rcv1_buff[i] = 0;
         rcv2_buff[i] = 0;
         rcv3_buff[i] = 0;
         rcv4_buff[i] = 0;
      }
   }

#endif

#if defined(__CATALINA_RECEIVER) && defined(__CATALINA_SENDER)

   ACQUIRE(lock);
   printf("SENDER AND RECEIVER (Clock = %lu Hz)\n\n", _clockfreq());
   printf("Press S to start sender cog or any other key to start receiver cog\n");
   printf("then do the opposite in the other program (note that the receiver\n");
   printf("should be stated first, then start the sender within 3 seconds)\n");
   RELEASE(lock);
   ch = k_wait();

   while (1) {

      if (toupper(ch) == 'S') {

         ACQUIRE(lock);
         printf("Starting sender cog ...\n");
         RELEASE(lock);
         send_cog = _cogstart_C(&sender, send_buff, send_stack, STACK_SIZE);

         // give the sender a chance to send (1 second is generous!)
         _waitms(1000);
         ACQUIRE(lock);
         printf("... done\n");
         RELEASE(lock);

         // cancel the sender (we restart it again for each transfer)
         _cogstop(send_cog);

      }
      else {

         ACQUIRE(lock);
         printf("Starting receiver cog ...\n");
         RELEASE(lock);

         // set a start time for the receiver cog to use in the sync function
         //start = _cnt() + _clockfreq(); // set start time for +1 seconds

         // start receiver cog
         rcv1_cog = _cogstart_C(&receiver, rcv1_buff, rcv1_stack, STACK_SIZE);
         ACQUIRE(lock);
         printf("... done\n");
         RELEASE(lock);

         // give the receiver a chance to receive (3 seconds is generous!)

         // wait till receivers completes
         {
            int count = 0;
            while ((rcv1_buff[NUM_LONGS - 1] == 0) && (count < 3)) {
               _waitms(1000);
               count++;
            }
         }


         // terminate the receiver cog (we restart it again for each transfer)
         _cogstop(rcv1_cog);

         // check the results
         ACQUIRE(lock);
         printf("\nChecking data ...\n");
         for (i = 0; i < NUM_LONGS; i++) {
            // check receiver 1 got the correct data
            if (send_buff[i] != rcv1_buff[i]) {
               printf("send[%3d]=0x%08X != rcv1[%3d]=0x%08X\n",  
                      i, send_buff[i], i, rcv1_buff[i]);
               _waitms(5);
#ifdef TOTAL_ERRORS_ONLY
               // if only counting total errors, we are done - we don't 
               // report each mismatch
               errors++;
               break;
#endif
            }
         }
         printf("... done\n");
         // update and print statistics
         transfers++;
         printf("Total errors = %d (from %d transfers)\n\n", errors, transfers); 
         RELEASE(lock);
      }

      // re-initialize the arrays for the next transfer
      for (i = 0; i < NUM_LONGS; i++) {
         rcv1_buff[i] = 0;
      }

      // now swap roles
      if (toupper(ch) == 'S') {
         ch = 'R';
      }
      else {
         _waitms(3000);
         ch = 'S';
      }
   }

#endif

}
